# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("mapi32.dll")
prototypes = \
    {
        #
        'OpenTnefStream': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="IStream"), SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="IMessage"), SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimTypeBottom(label="ITnef"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpvSupport", "lpStream", "lpszStreamName", "ulFlags", "lpMessage", "wKeyVal", "lppTNEF"]),
        #
        'OpenTnefStreamEx': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="IStream"), SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="IMessage"), SimTypeShort(signed=False, label="UInt16"), SimTypeBottom(label="IAddrBook"), SimTypePointer(SimTypeBottom(label="ITnef"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpvSupport", "lpStream", "lpszStreamName", "ulFlags", "lpMessage", "wKeyVal", "lpAdressBook", "lppTNEF"]),
        #
        'GetTnefStreamCodepage': SimTypeFunction([SimTypeBottom(label="IStream"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpStream", "lpulCodepage", "lpulSubCodepage"]),
        #
        'OpenIMsgSession': SimTypeFunction([SimTypeBottom(label="IMalloc"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpMalloc", "ulFlags", "lppMsgSess"]),
        #
        'CloseIMsgSession': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["lpMsgSess"]),
        #
        'OpenIMsgOnIStg': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lpObject", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer"]), offset=0), SimTypeBottom(label="IMalloc"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="IStorage"), SimTypePointer(SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="IMessage")], SimTypeBottom(label="Void"), arg_names=["ulCallerData", "lpMessage"]), offset=0), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IMessage"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpMsgSess", "lpAllocateBuffer", "lpAllocateMore", "lpFreeBuffer", "lpMalloc", "lpMapiSup", "lpStg", "lpfMsgCallRelease", "ulCallerData", "ulFlags", "lppMsg"]),
        #
        'GetAttribIMsgOnIStg': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeRef("SPropTagArray", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SPropAttrArray", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpObject", "lpPropTagArray", "lppPropAttrArray"]),
        #
        'SetAttribIMsgOnIStg': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeRef("SPropTagArray", SimStruct), offset=0), SimTypePointer(SimTypeRef("SPropAttrArray", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SPropProblemArray", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpObject", "lpPropTags", "lpPropAttrs", "lppPropProblems"]),
        #
        'MapStorageSCode': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["StgSCode"]),
        #
        'CreateIProp': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lpObject", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="IPropData"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpInterface", "lpAllocateBuffer", "lpAllocateMore", "lpFreeBuffer", "lpvReserved", "lppPropData"]),
        #
        'MAPIInitIdle': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpvReserved"]),
        #
        'MAPIDeinitIdle': SimTypeFunction([], SimTypeBottom(label="Void")),
        #
        'FtgRegisterIdleRoutine': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeShort(signed=True, label="Int16"), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["lpfnIdle", "lpvIdleParam", "priIdle", "csecIdle", "iroIdle"]),
        #
        'DeregisterIdleRoutine': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["ftg"]),
        #
        'EnableIdleRoutine': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeBottom(label="Void"), arg_names=["ftg", "fEnable"]),
        #
        'ChangeIdleRoutine': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeShort(signed=True, label="Int16"), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16"), SimTypeShort(signed=False, label="UInt16")], SimTypeBottom(label="Void"), arg_names=["ftg", "lpfnIdle", "lpvIdleParam", "priIdle", "csecIdle", "iroIdle", "ircIdle"]),
        #
        'MAPIGetDefaultMalloc': SimTypeFunction([], SimTypeBottom(label="IMalloc")),
        #
        'OpenStreamOnFile': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer"]), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypePointer(SimTypeBottom(label="IStream"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpAllocateBuffer", "lpFreeBuffer", "ulFlags", "lpszFileName", "lpszPrefix", "lppStream"]),
        #
        'PropCopyMore': SimTypeFunction([SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lpObject", "lppBuffer"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpSPropValueDest", "lpSPropValueSrc", "lpfAllocMore", "lpvObject"]),
        #
        'UlPropSize': SimTypeFunction([SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpSPropValue"]),
        #
        'FEqualNames': SimTypeFunction([SimTypePointer(SimTypeRef("MAPINAMEID", SimStruct), offset=0), SimTypePointer(SimTypeRef("MAPINAMEID", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpName1", "lpName2"]),
        #
        'FPropContainsProp': SimTypeFunction([SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpSPropValueDst", "lpSPropValueSrc", "ulFuzzyLevel"]),
        #
        'FPropCompareProp': SimTypeFunction([SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpSPropValue1", "ulRelOp", "lpSPropValue2"]),
        #
        'LPropCompareProp': SimTypeFunction([SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpSPropValueA", "lpSPropValueB"]),
        #
        'HrAddColumns': SimTypeFunction([SimTypeBottom(label="IMAPITable"), SimTypePointer(SimTypeRef("SPropTagArray", SimStruct), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer"]), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lptbl", "lpproptagColumnsNew", "lpAllocateBuffer", "lpFreeBuffer"]),
        #
        'HrAddColumnsEx': SimTypeFunction([SimTypeBottom(label="IMAPITable"), SimTypePointer(SimTypeRef("SPropTagArray", SimStruct), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer"]), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lptbl", "lpproptagColumnsNew", "lpAllocateBuffer", "lpFreeBuffer", "lpfnFilterColumns"]),
        #
        'HrAllocAdviseSink': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("NOTIFICATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpvContext", "cNotification", "lpNotifications"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="IMAPIAdviseSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpfnCallback", "lpvContext", "lppAdviseSink"]),
        #
        'HrThisThreadAdviseSink': SimTypeFunction([SimTypeBottom(label="IMAPIAdviseSink"), SimTypePointer(SimTypeBottom(label="IMAPIAdviseSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpAdviseSink", "lppAdviseSink"]),
        #
        'HrDispatchNotifications': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["ulFlags"]),
        #
        'BuildDisplayTable': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lpObject", "lppBuffer"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpBuffer"]), offset=0), SimTypeBottom(label="IMalloc"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DTPAGE", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IMAPITable"), offset=0), SimTypePointer(SimTypeBottom(label="ITableData"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpAllocateBuffer", "lpAllocateMore", "lpFreeBuffer", "lpMalloc", "hInstance", "cPages", "lpPage", "ulFlags", "lppTable", "lppTblData"]),
        #
        'ScCountNotifications': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("NOTIFICATION", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cNotifications", "lpNotifications", "lpcb"]),
        #
        'ScCopyNotifications': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("NOTIFICATION", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cNotification", "lpNotifications", "lpvDst", "lpcb"]),
        #
        'ScRelocNotifications': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("NOTIFICATION", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cNotification", "lpNotifications", "lpvBaseOld", "lpvBaseNew", "lpcb"]),
        #
        'ScCountProps': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cValues", "lpPropArray", "lpcb"]),
        #
        'LpValFindProp': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0)], SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), arg_names=["ulPropTag", "cValues", "lpPropArray"]),
        #
        'ScCopyProps': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cValues", "lpPropArray", "lpvDst", "lpcb"]),
        #
        'ScRelocProps': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cValues", "lpPropArray", "lpvBaseOld", "lpvBaseNew", "lpcb"]),
        #
        'ScDupPropset': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbSize", "lppBuffer"]), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cValues", "lpPropArray", "lpAllocateBuffer", "lppPropArray"]),
        #
        'UlAddRef': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpunk"]),
        #
        'UlRelease': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpunk"]),
        #
        'HrGetOneProp': SimTypeFunction([SimTypeBottom(label="IMAPIProp"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpMapiProp", "ulPropTag", "lppProp"]),
        #
        'HrSetOneProp': SimTypeFunction([SimTypeBottom(label="IMAPIProp"), SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpMapiProp", "lpProp"]),
        #
        'FPropExists': SimTypeFunction([SimTypeBottom(label="IMAPIProp"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpMapiProp", "ulPropTag"]),
        #
        'PpropFindProp': SimTypeFunction([SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeRef("SPropValue", SimStruct), offset=0), arg_names=["lpPropArray", "cValues", "ulPropTag"]),
        #
        'FreePadrlist': SimTypeFunction([SimTypePointer(SimTypeRef("ADRLIST", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["lpAdrlist"]),
        #
        'FreeProws': SimTypeFunction([SimTypePointer(SimTypeRef("SRowSet", SimStruct), offset=0)], SimTypeBottom(label="Void"), arg_names=["lpRows"]),
        #
        'HrQueryAllRows': SimTypeFunction([SimTypeBottom(label="IMAPITable"), SimTypePointer(SimTypeRef("SPropTagArray", SimStruct), offset=0), SimTypePointer(SimTypeRef("SRestriction", SimStruct), offset=0), SimTypePointer(SimTypeRef("SSortOrderSet", SimStruct), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimTypeRef("SRowSet", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpTable", "lpPropTags", "lpRestriction", "lpSortOrderSet", "crowsMax", "lppRows"]),
        #
        'SzFindCh': SimTypeFunction([SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypeShort(signed=False, label="UInt16")], SimTypePointer(SimTypeChar(label="SByte"), offset=0), arg_names=["lpsz", "ch"]),
        #
        'SzFindLastCh': SimTypeFunction([SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypeShort(signed=False, label="UInt16")], SimTypePointer(SimTypeChar(label="SByte"), offset=0), arg_names=["lpsz", "ch"]),
        #
        'SzFindSz': SimTypeFunction([SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypePointer(SimTypeChar(label="SByte"), offset=0), arg_names=["lpsz", "lpszKey"]),
        #
        'UFromSz': SimTypeFunction([SimTypePointer(SimTypeChar(label="SByte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpsz"]),
        #
        'ScUNCFromLocalPath': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszLocal", "lpszUNC", "cchUNC"]),
        #
        'ScLocalPathFromUNC': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszUNC", "lpszLocal", "cchLocal"]),
        #
        'FtAddFt': SimTypeFunction([SimTypeRef("FILETIME", SimStruct), SimTypeRef("FILETIME", SimStruct)], SimTypeRef("FILETIME", SimStruct), arg_names=["ftAddend1", "ftAddend2"]),
        #
        'FtMulDwDw': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeRef("FILETIME", SimStruct), arg_names=["ftMultiplicand", "ftMultiplier"]),
        #
        'FtMulDw': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeRef("FILETIME", SimStruct)], SimTypeRef("FILETIME", SimStruct), arg_names=["ftMultiplier", "ftMultiplicand"]),
        #
        'FtSubFt': SimTypeFunction([SimTypeRef("FILETIME", SimStruct), SimTypeRef("FILETIME", SimStruct)], SimTypeRef("FILETIME", SimStruct), arg_names=["ftMinuend", "ftSubtrahend"]),
        #
        'FtNegFt': SimTypeFunction([SimTypeRef("FILETIME", SimStruct)], SimTypeRef("FILETIME", SimStruct), arg_names=["ft"]),
        #
        'ScCreateConversationIndex': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cbParent", "lpbParent", "lpcbConvIndex", "lppbConvIndex"]),
        #
        'WrapStoreEntryID': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="SByte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("ENTRYID", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("ENTRYID", SimStruct), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ulFlags", "lpszDLLName", "cbOrigEntry", "lpOrigEntry", "lpcbWrappedEntry", "lppWrappedEntry"]),
        #
        'RTFSync': SimTypeFunction([SimTypeBottom(label="IMessage"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpMessage", "ulFlags", "lpfMessageUpdated"]),
        #
        'WrapCompressedRTFStream': SimTypeFunction([SimTypeBottom(label="IStream"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IStream"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpCompressedRTFStream", "ulFlags", "lpUncompressedRTFStream"]),
        #
        'HrIStorageFromStream': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IStorage"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpUnkIn", "lpInterface", "ulFlags", "lppStorageOut"]),
        #
        'ScInitMapiUtil': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["ulFlags"]),
        #
        'DeinitMapiUtil': SimTypeFunction([], SimTypeBottom(label="Void")),
        #
        'MAPIFreeBuffer': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pv"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations

import logging

import claripy

from angr.misc.ux import once
from angr.misc.picklable_lock import PicklableLock
from angr.state_plugins.sim_action import SimActionConstraint
from angr.state_plugins.sim_action_object import SimActionObject
from .base import ExplorationTechnique

l = logging.getLogger(__name__)


def ast_weight(ast, memo=None):
    if isinstance(ast, SimActionObject):
        ast = ast.ast
    if not isinstance(ast, claripy.ast.Base):
        return 0

    if memo is None:
        memo = {}

    result = memo.get(ast.hash(), None)
    if result is not None:
        return result

    result = 1 + sum(ast_weight(arg, memo) for arg in ast.args)
    memo[ast.hash()] = result
    return result


class Suggestions(ExplorationTechnique):
    """
    An exploration technique which analyzes failure cases and logs suggestions for how to mitigate them in future
    analyses.
    """

    def __init__(self):
        super().__init__()
        self.suggested = set()
        self.lock = PicklableLock()

    def step(self, simgr, stash="active", **kwargs):
        simgr.step(stash=stash, **kwargs)

        for state in simgr.stashes.get("interrupted", []):
            if id(state) in self.suggested:
                continue
            self.suggested.add(id(state))

            try:
                event = state.history.events[-1]
            except IndexError:
                continue
            if event.type != "insufficient_resources":
                continue

            with self.lock:  # do not interleave logs
                self.report(state, event)

        return simgr

    @staticmethod
    def report(state, event):
        if once("suggestion_technique"):
            l.warning("Some of your states hit a resource limit. set logger %s to INFO for suggestions.", __name__)
            l.info("Create your simulation manager with `suggestions=False` to disable this.")

        if event.objects["type"] is claripy.errors.ClaripySolverInterruptError:
            if event.objects["reason"][0] == "timeout":
                limit_number = state.solver._solver.timeout
                limit_kind = f"hit a solver timeout of {limit_number} ms."
                limit_minimum = 60 * 1000
            elif event.objects["reason"][0] == "max. memory exceeded":
                limit_number = state.solver._solver.max_memory
                limit_kind = f"hit a solver memory limit of {limit_number} MB."
                limit_minimum = 1024
            else:
                limit_number = None
                limit_kind = "hit an unknown resource limit. are you manually mucking with the z3 backend?"
                limit_minimum = None
            l.info("%s %s", state, limit_kind)
            if limit_number is not None and limit_minimum is not None and limit_number < limit_minimum:
                l.info("The minimum recommended limit is %s. Consider turning it up?", limit_minimum)

            log = []
            for history in state.history.lineage:
                for constraint_event in history.recent_events:
                    if isinstance(constraint_event, SimActionConstraint):
                        constraint = constraint_event.constraint.ast
                        if constraint is history.jump_guard:
                            src_addr = history.jump_source
                            dst_addr = history.jump_target
                            transition_type = "jumping"
                        else:
                            if constraint_event.sim_procedure is None:
                                src_addr = constraint_event.ins_addr
                            else:
                                src_addr = constraint_event.sim_procedure.addr
                            block = state.block(src_addr, num_inst=2)
                            try:
                                dst_addr = block.instruction_addrs[1]
                            except IndexError:
                                dst_addr = history.jump_target
                            transition_type = "stepping"
                        if type(dst_addr) is int:
                            dst_addr = claripy.BVV(dst_addr, state.arch.bits)
                        log.append((constraint, ast_weight(constraint), src_addr, dst_addr, transition_type, len(log)))

            log.sort(key=lambda t: t[1], reverse=True)
            max_delta_idx = None
            if len(log) > 1:
                deltas = [b[1] - a[1] for a, b in zip(log, log[1:])]
                max_delta_idx, max_delta = max(enumerate(deltas))
                if max_delta < 2**10:
                    max_delta_idx = None
            if max_delta_idx is None and log and log[0][1] >= 2**16:
                for max_delta_idx, t in enumerate(log):
                    if t[1] < 2**16:
                        max_delta_idx -= 1
                        break
            if max_delta_idx is not None:
                l.info(
                    "%d constraint%s abnormally complex.", max_delta_idx + 1, "s are" if max_delta_idx > 0 else " is"
                )
                descriptions = []
                for t in sorted(log[: max_delta_idx + 1], key=lambda t: t[5]):
                    if max_delta_idx < 10:
                        l.info(
                            "...generated %s from %s to %s",
                            t[4],
                            state.project.loader.describe_addr(t[2]),
                            state.project.loader.describe_addr(t[3].args[0]) if t[3].op == "BVV" else "<symbol>",
                        )
                    descriptions.extend(state.solver.describe_variables(t[0]))
                descriptions = set(descriptions)
                seen_apis = set()
                for description in descriptions:
                    if description[0] == "api":
                        if description[1] not in seen_apis:
                            seen_apis.add(description[1])
                            l.info("...using variables originating in %s (hook it?)", description[1])
                    elif description[0] == "mem":
                        if "##mem" not in seen_apis:
                            seen_apis.add("##mem")
                            l.info("...using unconstrained memory (add ZERO_FILL_UNCONSTRAINED_MEMORY?)")
                    elif description[0] == "reg":
                        if "##reg" not in seen_apis:
                            seen_apis.add("##reg")
                            l.info("...using unconstrained registers (add ZERO_FILL_UNCONSTRAINED_REGISTERS?)")
                    elif description[0] == "file":
                        api = "file##" + description[1]
                        if api not in seen_apis:
                            seen_apis.add(api)
                            l.info("...using variables from file %s", description[1])
                    else:
                        l.info("...using uncategorized variable %s", description)



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("wldp.dll")
prototypes = \
    {
        #
        'WldpGetLockdownPolicy': SimTypeFunction([SimTypePointer(SimTypeRef("WLDP_HOST_INFORMATION", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hostInformation", "lockdownState", "lockdownFlags"]),
        #
        'WldpIsClassInApprovedList': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeRef("WLDP_HOST_INFORMATION", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["classID", "hostInformation", "isApproved", "optionalFlags"]),
        #
        'WldpSetDynamicCodeTrust': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["fileHandle"]),
        #
        'WldpIsDynamicCodePolicyEnabled': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["isEnabled"]),
        #
        'WldpQueryDynamicCodeTrust': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["fileHandle", "baseImage", "imageSize"]),
        #
        'WldpQueryDeviceSecurityInformation': SimTypeFunction([SimTypePointer(SimTypeRef("WLDP_DEVICE_SECURITY_INFORMATION", SimStruct), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["information", "informationLength", "returnLength"]),
        #
        'WldpCanExecuteFile': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="WLDP_EXECUTION_EVALUATION_OPTIONS"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WLDP_EXECUTION_POLICY"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["host", "options", "fileHandle", "auditInfo", "result"]),
        #
        'WldpCanExecuteBuffer': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="WLDP_EXECUTION_EVALUATION_OPTIONS"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WLDP_EXECUTION_POLICY"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["host", "options", "buffer", "bufferSize", "auditInfo", "result"]),
        #
        'WldpCanExecuteStream': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="WLDP_EXECUTION_EVALUATION_OPTIONS"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WLDP_EXECUTION_POLICY"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["host", "options", "stream", "auditInfo", "result"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations

import abc
from typing import Generic, TypeVar


import angr

StateType = TypeVar("StateType")
ResultType = TypeVar("ResultType")
DataType_co = TypeVar("DataType_co", covariant=True)


class SimEngine(Generic[StateType, ResultType], metaclass=abc.ABCMeta):
    """
    A SimEngine is a type which understands how to perform execution on a state.
    """

    state: StateType

    def __init__(self, project: angr.Project):
        self.project = project
        self.arch = self.project.arch

    def __getstate__(self):
        return (self.project,)

    def __setstate__(self, state):
        self.project = state[0]



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("mf.dll")
prototypes = \
    {
        #
        'MFCreateMediaSession': SimTypeFunction([SimTypeBottom(label="IMFAttributes"), SimTypePointer(SimTypeBottom(label="IMFMediaSession"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pConfiguration", "ppMediaSession"]),
        #
        'MFCreatePMPMediaSession': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="IMFAttributes"), SimTypePointer(SimTypeBottom(label="IMFMediaSession"), offset=0), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwCreationFlags", "pConfiguration", "ppMediaSession", "ppEnablerActivate"]),
        #
        'MFCreateTopology': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFTopology"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppTopo"]),
        #
        'MFCreateTopologyNode': SimTypeFunction([SimTypeInt(signed=False, label="MF_TOPOLOGY_TYPE"), SimTypePointer(SimTypeBottom(label="IMFTopologyNode"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["NodeType", "ppNode"]),
        #
        'MFGetTopoNodeCurrentType': SimTypeFunction([SimTypeBottom(label="IMFTopologyNode"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="IMFMediaType"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pNode", "dwStreamIndex", "fOutput", "ppType"]),
        #
        'MFGetService': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["punkObject", "guidService", "riid", "ppvObject"]),
        #
        'MFCreatePresentationClock': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFPresentationClock"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppPresentationClock"]),
        #
        'MFRequireProtectedEnvironment': SimTypeFunction([SimTypeBottom(label="IMFPresentationDescriptor")], SimTypeInt(signed=True, label="Int32"), arg_names=["pPresentationDescriptor"]),
        #
        'MFCreateSimpleTypeHandler': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFMediaTypeHandler"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppHandler"]),
        #
        'MFShutdownObject': SimTypeFunction([SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["pUnk"]),
        #
        'MFCreateAudioRenderer': SimTypeFunction([SimTypeBottom(label="IMFAttributes"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAudioAttributes", "ppSink"]),
        #
        'MFCreateAudioRendererActivate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppActivate"]),
        #
        'MFCreateVideoRendererActivate': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwndVideo", "ppActivate"]),
        #
        'MFCreateMPEG4MediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypeBottom(label="IMFMediaType"), SimTypeBottom(label="IMFMediaType"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIByteStream", "pVideoMediaType", "pAudioMediaType", "ppIMediaSink"]),
        #
        'MFCreate3GPMediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypeBottom(label="IMFMediaType"), SimTypeBottom(label="IMFMediaType"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIByteStream", "pVideoMediaType", "pAudioMediaType", "ppIMediaSink"]),
        #
        'MFCreateMP3MediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pTargetByteStream", "ppMediaSink"]),
        #
        'MFCreateAC3MediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypeBottom(label="IMFMediaType"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pTargetByteStream", "pAudioMediaType", "ppMediaSink"]),
        #
        'MFCreateADTSMediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypeBottom(label="IMFMediaType"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pTargetByteStream", "pAudioMediaType", "ppMediaSink"]),
        #
        'MFCreateMuxSink': SimTypeFunction([SimTypeBottom(label="Guid"), SimTypeBottom(label="IMFAttributes"), SimTypeBottom(label="IMFByteStream"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["guidOutputSubType", "pOutputAttributes", "pOutputByteStream", "ppMuxSink"]),
        #
        'MFCreateFMPEG4MediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypeBottom(label="IMFMediaType"), SimTypeBottom(label="IMFMediaType"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIByteStream", "pVideoMediaType", "pAudioMediaType", "ppIMediaSink"]),
        #
        'MFCreateTopoLoader': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFTopoLoader"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppObj"]),
        #
        'MFCreateSampleGrabberSinkActivate': SimTypeFunction([SimTypeBottom(label="IMFMediaType"), SimTypeBottom(label="IMFSampleGrabberSinkCallback"), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIMFMediaType", "pIMFSampleGrabberSinkCallback", "ppIActivate"]),
        #
        'MFCreateStandardQualityManager': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFQualityManager"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppQualityManager"]),
        #
        'MFCreateSequencerSource': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypePointer(SimTypeBottom(label="IMFSequencerSource"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pReserved", "ppSequencerSource"]),
        #
        'MFCreateSequencerSegmentOffset': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeLongLong(signed=True, label="Int64"), SimTypePointer(SimTypeRef("PROPVARIANT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwId", "hnsOffset", "pvarSegmentOffset"]),
        #
        'MFCreateAggregateSource': SimTypeFunction([SimTypeBottom(label="IMFCollection"), SimTypePointer(SimTypeBottom(label="IMFMediaSource"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSourceCollection", "ppAggSource"]),
        #
        'MFCreateCredentialCache': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFNetCredentialCache"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppCache"]),
        #
        'MFCreateProxyLocator': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IPropertyStore"), SimTypePointer(SimTypeBottom(label="IMFNetProxyLocator"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszProtocol", "pProxyConfig", "ppProxyLocator"]),
        #
        'MFCreateNetSchemePlugin': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["riid", "ppvHandler"]),
        #
        'MFCreatePMPServer': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IMFPMPServer"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["dwCreationFlags", "ppPMPServer"]),
        #
        'MFCreateRemoteDesktopPlugin': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFRemoteDesktopPlugin"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppPlugin"]),
        #
        'CreateNamedPropertyStore': SimTypeFunction([SimTypePointer(SimTypeBottom(label="INamedPropertyStore"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppStore"]),
        #
        'MFCreateSampleCopierMFT': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFTransform"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppCopierMFT"]),
        #
        'MFCreateTranscodeProfile': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFTranscodeProfile"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppTranscodeProfile"]),
        #
        'MFCreateTranscodeTopology': SimTypeFunction([SimTypeBottom(label="IMFMediaSource"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IMFTranscodeProfile"), SimTypePointer(SimTypeBottom(label="IMFTopology"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrc", "pwszOutputFilePath", "pProfile", "ppTranscodeTopo"]),
        #
        'MFCreateTranscodeTopologyFromByteStream': SimTypeFunction([SimTypeBottom(label="IMFMediaSource"), SimTypeBottom(label="IMFByteStream"), SimTypeBottom(label="IMFTranscodeProfile"), SimTypePointer(SimTypeBottom(label="IMFTopology"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrc", "pOutputStream", "pProfile", "ppTranscodeTopo"]),
        #
        'MFTranscodeGetAudioOutputAvailableTypes': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="IMFAttributes"), SimTypePointer(SimTypeBottom(label="IMFCollection"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["guidSubType", "dwMFTFlags", "pCodecConfig", "ppAvailableTypes"]),
        #
        'MFCreateTranscodeSinkActivate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppActivate"]),
        #
        'MFEnumDeviceSources': SimTypeFunction([SimTypeBottom(label="IMFAttributes"), SimTypePointer(SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAttributes", "pppSourceActivate", "pcSourceActivate"]),
        #
        'MFCreateDeviceSource': SimTypeFunction([SimTypeBottom(label="IMFAttributes"), SimTypePointer(SimTypeBottom(label="IMFMediaSource"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAttributes", "ppSource"]),
        #
        'MFCreateDeviceSourceActivate': SimTypeFunction([SimTypeBottom(label="IMFAttributes"), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAttributes", "ppActivate"]),
        #
        'MFCreateProtectedEnvironmentAccess': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFProtectedEnvironmentAccess"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppAccess"]),
        #
        'MFLoadSignedLibrary': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="IMFSignedLibrary"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszName", "ppLib"]),
        #
        'MFGetSystemId': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFSystemId"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppId"]),
        #
        'MFGetLocalId': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["verifier", "size", "id"]),
        #
        'MFCreateASFContentInfo': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFASFContentInfo"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppIContentInfo"]),
        #
        'MFCreateASFIndexer': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFASFIndexer"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppIIndexer"]),
        #
        'MFCreateASFIndexerByteStream': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimTypeBottom(label="IMFByteStream"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIContentByteStream", "cbIndexStartOffset", "pIIndexByteStream"]),
        #
        'MFCreateASFSplitter': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFASFSplitter"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppISplitter"]),
        #
        'MFCreateASFProfile': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFASFProfile"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppIProfile"]),
        #
        'MFCreateASFProfileFromPresentationDescriptor': SimTypeFunction([SimTypeBottom(label="IMFPresentationDescriptor"), SimTypePointer(SimTypeBottom(label="IMFASFProfile"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIPD", "ppIProfile"]),
        #
        'MFCreatePresentationDescriptorFromASFProfile': SimTypeFunction([SimTypeBottom(label="IMFASFProfile"), SimTypePointer(SimTypeBottom(label="IMFPresentationDescriptor"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIProfile", "ppIPD"]),
        #
        'MFCreateASFMultiplexer': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IMFASFMultiplexer"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppIMultiplexer"]),
        #
        'MFCreateASFStreamSelector': SimTypeFunction([SimTypeBottom(label="IMFASFProfile"), SimTypePointer(SimTypeBottom(label="IMFASFStreamSelector"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIASFProfile", "ppSelector"]),
        #
        'MFCreateASFMediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIByteStream", "ppIMediaSink"]),
        #
        'MFCreateASFMediaSinkActivate': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IMFASFContentInfo"), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszFileName", "pContentInfo", "ppIActivate"]),
        #
        'MFCreateWMVEncoderActivate': SimTypeFunction([SimTypeBottom(label="IMFMediaType"), SimTypeBottom(label="IPropertyStore"), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pMediaType", "pEncodingConfigurationProperties", "ppActivate"]),
        #
        'MFCreateWMAEncoderActivate': SimTypeFunction([SimTypeBottom(label="IMFMediaType"), SimTypeBottom(label="IPropertyStore"), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pMediaType", "pEncodingConfigurationProperties", "ppActivate"]),
        #
        'MFCreateASFStreamingMediaSink': SimTypeFunction([SimTypeBottom(label="IMFByteStream"), SimTypePointer(SimTypeBottom(label="IMFMediaSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pIByteStream", "ppIMediaSink"]),
        #
        'MFCreateASFStreamingMediaSinkActivate': SimTypeFunction([SimTypeBottom(label="IMFActivate"), SimTypeBottom(label="IMFASFContentInfo"), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pByteStreamActivate", "pContentInfo", "ppIActivate"]),
        #
        'MFCreateVideoRenderer': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["riidRenderer", "ppVideoRenderer"]),
        #
        'MFCreateEncryptedMediaExtensionsStoreActivate': SimTypeFunction([SimTypeBottom(label="IMFPMPHostApp"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="IMFActivate"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pmpHost", "objectStream", "classId", "activate"]),
    }

lib.set_prototypes(prototypes)
